import base64


class CommonUtil:
    @staticmethod
    def decode_response(response: dict = None):
        if response is None:
            response = {}
        if "data" in response:
            array_decode = ["external_ids", "content", "names"]
            if isinstance(response["data"], list):
                for item in response["data"]:
                    for arr in array_decode:
                        if arr in item:
                            item[arr] = CommonUtil.decode(arr, item[arr])
            else:
                for arr in array_decode:
                    if arr in response["data"]:
                        response["data"][arr] = CommonUtil.decode(arr, response["data"][arr])
        return response

    @staticmethod
    def decode(name: str, data):
        if isinstance(data, list):
            decoded = []
            if name == "external_ids" and len(data) >= 6 \
                    and (base64.b64decode(data[0]) == b"SignedChain"
                         or base64.b64decode(data[0]) == b"SignedEntry") \
                    and base64.b64decode(data[1]).hex() == "01":
                for idx, val in enumerate(data):
                    if idx == 1:
                        decoded.append("0x" + base64.b64decode(val).hex())
                    elif idx == 4:
                        decoded.append(base64.b64decode(val).hex())
                    else:
                        decoded.append("".join(chr(x) for x in base64.b64decode(val)))
            else:
                decoded = ["".join(chr(x) for x in base64.b64decode(val)) for val in data]
            return decoded
        else:
            return "".join(chr(x) for x in base64.b64decode(data))

    @staticmethod
    def base64_encode(value):
        if isinstance(value, str):
            return "".join(chr(x) for x in base64.b64encode(value.encode()))
        return "".join(chr(x) for x in base64.b64encode(value))



import datetime

import validators

import factom_sdk.utils.consts
from factom_sdk.request_handler.request_handler import RequestHandler
from factom_sdk.utils.common_util import CommonUtil
from factom_sdk.utils.key_common import KeyCommon
from factom_sdk.utils.utils import Utils
from factom_sdk.utils.validate_signature_util import ValidateSignatureUtil


class EntriesClient:
    def __init__(self, base_url: str, app_id: str, app_key: str, automatic_signing: bool):
        self.request_handler = RequestHandler(base_url, app_id, app_key)
        self.automatic_signing = automatic_signing

    def get(self, chain_id: str, entry_hash: str, signature_validation=None):
        """Gets information about a specific entry on Connect.

        Args:
            chain_id (str): The chain identifier.
            entry_hash (str): The SHA256 hash of the entry.
            signature_validation (bool | custom function)

        Returns:
            Entry object.
        """
        if not chain_id:
            raise Exception("chain_id is required.")
        if not entry_hash:
            raise Exception("entry_hash is required.")
        response = self.request_handler.get("/".join([factom_sdk.utils.consts.CHAINS_URL, chain_id,
                                                      factom_sdk.utils.consts.ENTRIES_URL, entry_hash]))
        if not callable(signature_validation) and not isinstance(signature_validation, bool):
            signature_validation = True
        if signature_validation and isinstance(signature_validation, bool):
            return {
                "entry": response,
                "status": ValidateSignatureUtil.validate_signature(response, False, self.request_handler)
            }
        elif callable(signature_validation):
            return {
                "entry": response,
                "status": signature_validation(response)
            }
        return response

    def create(self, chain_id: str, content: str, external_ids: list = None,
               signer_private_key: str = "", signer_chain_id: str = "", callback_url: str = "",
               callback_stages: list = None):
        """Creates a new entry for the selected chain.

        Args:
            chain_id (str): The chain identifier.
            content (str): This is the data that will be stored directly on the blockchain.
            external_ids (:obj:`list`, optional): Tags that can be used to identify your entry.
            signer_private_key (:obj:`str`, optional): base58 string in Idsec format. This parameter is optional for
            creating an unsigned entry.
            signer_chain_id (:obj:`str`, optional): The chain id of the signer identity. This parameter is optional for
            creating an unsigned entry.
            callback_url (:obj:`str`, optional): The URL where you would like to receive the callback from Connect.
            callback_stages (:obj:`list`, optional): The immutability stages you would like to be notified about.
            This list can include any or all of the three stages: `replicated`, `factom`, and `anchored`.

        Returns:
            Entry created object.
        """
        if callback_stages is None:
            callback_stages = []
        if external_ids is None:
            external_ids = []
        if not chain_id:
            raise Exception("chain_id is required.")
        if self.automatic_signing:
            if not isinstance(external_ids, list):
                raise Exception("external_ids must be an array.")
            if not signer_private_key:
                raise Exception("signer_private_key is required.")
            if not KeyCommon.validate_checksum(signer_private_key):
                raise Exception("signer_private_key is invalid.")
            if not signer_chain_id:
                raise Exception("signer_chain_id is required.")
        else:
            if not external_ids:
                raise Exception("at least 1 external_id is required.")
            if not isinstance(external_ids, list):
                raise Exception("external_ids must be an array.")
            if signer_private_key and not signer_chain_id:
                raise Exception("signer_chain_id is required when passing a signer_private_key.")
            if signer_private_key and not KeyCommon.validate_checksum(signer_private_key):
                raise Exception("signer_private_key is invalid.")
            if signer_chain_id and not signer_private_key:
                raise Exception("signer_private_key is required when passing a signer_chain_id.")
        if not content:
            raise Exception("content is required.")
        if callback_url and not validators.url(callback_url):
            raise Exception("callback_url is an invalid url format.")
        if not isinstance(callback_stages, list):
            raise Exception("callback_stages must be an array.")
        ids_base64 = []
        if self.automatic_signing:
            time_stamp = Utils.to_military_timezone_str(datetime.datetime.now(datetime.timezone.utc))
            message = signer_chain_id + content + time_stamp
            signature = KeyCommon.sign_content(signer_private_key, message)
            signer_public_key = KeyCommon.get_public_key_from_private_key(signer_private_key)
            ids_base64.append(CommonUtil.base64_encode("SignedEntry"))
            ids_base64.append(CommonUtil.base64_encode(bytes([0x01])))
            ids_base64.append(CommonUtil.base64_encode(signer_chain_id))
            ids_base64.append(CommonUtil.base64_encode(signer_public_key))
            ids_base64.append(signature)
            ids_base64.append(CommonUtil.base64_encode(time_stamp))
        for val in external_ids:
            ids_base64.append(CommonUtil.base64_encode(val))
        data = {
            "external_ids": ids_base64,
            "content": CommonUtil.base64_encode(content)
        }
        if callback_url:
            data["callback_url"] = callback_url
        if callback_stages:
            data["callback_stages"] = callback_stages
        return self.request_handler.post("/".join([factom_sdk.utils.consts.CHAINS_URL, chain_id,
                                                   factom_sdk.utils.consts.ENTRIES_URL]), data)

    def list(self, chain_id: str, limit: int = -1, offset: int = -1, stages: list = None):
        """Gets list of all entries contained on a specified chain.

        Args:
            chain_id (str): The chain identifier.
            limit (:obj:`int`, optional): The number of items you would like to return back in each stage.
            offset (:obj:`int`, optional): The offset parameter allows you to select which item you would like to start
            from when a list is returned from Connect.
            stages (:obj:`list`, optional): The immutability stages you want to restrict results to.
            You can choose any from `replicated`, `factom`, and `anchored`.

        Returns:
            List entry object.
        """
        if stages is None:
            stages = []
        if not chain_id:
            raise Exception("chain_id is required.")
        data = {}
        if not isinstance(limit, int):
            raise Exception("limit must be an integer.")
        if limit > -1:
            data["limit"] = limit
        if not isinstance(offset, int):
            raise Exception("offset must be an integer.")
        if offset > -1:
            data["offset"] = offset
        if not isinstance(stages, list):
            raise Exception("stages must be an array.")
        if stages:
            data["stages"] = ",".join(stages)
        return self.request_handler.get("/".join([factom_sdk.utils.consts.CHAINS_URL, chain_id,
                                                  factom_sdk.utils.consts.ENTRIES_URL]), data)

    def get_first(self, chain_id: str, signature_validation=None):
        """Retrieves the first entry that has been saved to this chain.

        Args:
            chain_id (str): The chain identifier.
            signature_validation (bool | custom function)

        Returns:
            Entry object.
        """
        if not chain_id:
            raise Exception("chain_id is required.")
        response = self.request_handler.get("/".join([factom_sdk.utils.consts.CHAINS_URL, chain_id,
                                                      factom_sdk.utils.consts.ENTRIES_URL,
                                                      factom_sdk.utils.consts.FIRST_URL]))
        if not callable(signature_validation) and not isinstance(signature_validation, bool):
            signature_validation = True
        if signature_validation and isinstance(signature_validation, bool):
            return {
                "entry": response,
                "status": ValidateSignatureUtil.validate_signature(response, False, self.request_handler)
            }
        elif callable(signature_validation):
            return {
                "entry": response,
                "status": signature_validation(response)
            }
        return response

    def get_last(self, chain_id: str, signature_validation=None):
        """Gets the last entry that has been saved to this chain.

        Args:
            chain_id (str): The chain identifier.
            signature_validation (bool | custom function)

        Returns:
            Entry object.
        """
        if not chain_id:
            raise Exception("chain_id is required.")
        response = self.request_handler.get("/".join([factom_sdk.utils.consts.CHAINS_URL, chain_id,
                                                      factom_sdk.utils.consts.ENTRIES_URL,
                                                      factom_sdk.utils.consts.LAST_URL]))
        if not callable(signature_validation) and not isinstance(signature_validation, bool):
            signature_validation = True
        if signature_validation and isinstance(signature_validation, bool):
            return {
                "entry": response,
                "status": ValidateSignatureUtil.validate_signature(response, False, self.request_handler)
            }
        elif callable(signature_validation):
            return {
                "entry": response,
                "status": signature_validation(response)
            }
        return response

    def search(self, chain_id: str, external_ids: list, limit: int = -1, offset: int = -1):
        """Finds all of the entries with `external_ids` that match what you entered.

        Args:
            chain_id (str): The chain identifier.
            external_ids (list): A list of external IDs.
            limit (:obj:`int`, optional): The number of items you would like to return back in each stage.
            offset (:obj:`int`, optional): The offset parameter allows you to select which item you would like to start
            from when a list is returned from Connect.

        Returns:
            List entry object.
        """
        if not chain_id:
            raise Exception("chain_id is required.")
        if not external_ids:
            raise Exception("at least 1 external_id is required.")
        if not isinstance(external_ids, list):
            raise Exception("external_ids must be an array.")
        url = "/".join([factom_sdk.utils.consts.CHAINS_URL, chain_id, factom_sdk.utils.consts.ENTRIES_URL,
                        factom_sdk.utils.consts.SEARCH_URL])
        if not isinstance(limit, int):
            raise Exception("limit must be an integer.")
        if limit > -1:
            url += "?limit=" + str(limit)
        if not isinstance(offset, int):
            raise Exception("offset must be an integer.")
        if offset > -1:
            if limit > -1:
                url += "&offset=" + str(offset)
            else:
                url += "?offset=" + str(offset)
        ids_base64 = [CommonUtil.base64_encode(val) for val in external_ids]
        data = {"external_ids": ids_base64}
        return self.request_handler.post(url, data)



import validators
import factom_sdk.utils.consts
from factom_sdk.utils.key_common import KeyCommon
from factom_sdk.request_handler.request_handler import RequestHandler


class IdentitiesKeyUtil:
    def __init__(self, base_url: str, app_id: str, app_key: str):
        self.request_handler = RequestHandler(base_url, app_id, app_key)

    def get(self, identity_chain_id: str, key: str):
        """Gets information about a specific public key for a given Identity, including the heights at which the key
        was activated and retired if applicable.

        Args:
            identity_chain_id (str): The unique identifier for the Identity that the key belongs to.
            key (str): The public key string to get information, which must be in base58 idpub format.

        Returns:
            Identity key object.
        """
        if not identity_chain_id:
            raise Exception("identity_chain_id is required.")
        if not key:
            raise Exception("key is required.")
        if not KeyCommon.validate_checksum(key):
            raise Exception("key is invalid.")
        return self.request_handler.get("/".join([factom_sdk.utils.consts.IDENTITY_URL, identity_chain_id,
                                                  factom_sdk.utils.consts.KEYS_STRING, key]))

    def list(self, identity_chain_id: str, limit: int = -1, offset: int = -1):
        """Returns all of the keys that were ever active for this Identity. Results are paginated.

        Args:
            identity_chain_id (str): The unique identifier of the identity chain whose keys are being requested.
            limit (:obj:`int`, optional): The maximum number of keys you would like to be returned.
            offset (:obj:`int`, optional): The key index (in number of keys from the first key) to start from in the
            list of all keys.

        Returns:
            List identity key object.
        """
        if not identity_chain_id:
            raise Exception("identity_chain_id is required.")
        data = {}
        if not isinstance(limit, int):
            raise Exception("limit must be an integer.")
        if limit > -1:
            data["limit"] = limit
        if not isinstance(offset, int):
            raise Exception("offset must be an integer.")
        if offset > -1:
            data["offset"] = offset
        return self.request_handler.get("/".join([factom_sdk.utils.consts.IDENTITY_URL, identity_chain_id,
                                                  factom_sdk.utils.consts.KEYS_STRING]), data)

    def replace(self, identity_chain_id: str, old_public_key: str, signer_private_key: str,
                new_public_key: str = None, callback_url: str = "", callback_stages: list = None):
        """Creates an entry in the Identity Chain for a key replacement.

        Args:
            identity_chain_id (str): The unique identifier of the identity chain being requested.
            old_public_key (str): base58 string in Idpub format. The public key to be retired and replaced.
            signer_private_key (str): base58 string in Idsec format. The private key to use to create the signature,
            which must be the same or higher priority than the public key to be replaced.
            new_public_key (:obj:`str`, optional): base58 string in Idpub format. The new public key to be activated
            and take its place.
            callback_url (:obj:`str`, optional): The URL you would like the callbacks to be sent to.
            callback_stages (:obj:`list`, optional): The immutability stages you'd like to be notified about. This list
            can include any or all of these three stages: `replicated`, `factom`, and `anchored`

        Returns:
            Replacement result object.
        """
        if callback_stages is None:
            callback_stages = []
        if not identity_chain_id:
            raise Exception("identity_chain_id is required.")
        if not old_public_key:
            raise Exception("old_public_key is required.")
        if not signer_private_key:
            raise Exception("signer_private_key is required.")
        key_pair = {}
        if new_public_key is not None:
            if not new_public_key:
                raise Exception("new_public_key is required.")
            new_key = new_public_key
        else:
            key_pair = KeyCommon.create_key_pair()
            new_key = key_pair["public_key"]
        if not KeyCommon.validate_checksum(old_public_key):
            raise Exception("old_public_key is an invalid public key.")
        if not KeyCommon.validate_checksum(signer_private_key):
            raise Exception("signer_private_key is invalid.")
        if not KeyCommon.validate_checksum(new_key):
            raise Exception("new_public_key is an invalid public key.")
        if not isinstance(callback_stages, list):
            raise Exception("callback_stages must be an array.")
        if callback_url and not validators.url(callback_url):
            raise Exception("callback_url is an invalid url format.")
        message = identity_chain_id + old_public_key + new_key
        signature = KeyCommon.sign_content(signer_private_key, message)
        signer_key = KeyCommon.get_public_key_from_private_key(signer_private_key)
        data = {
            "old_key": old_public_key,
            "new_key": new_key,
            "signature": signature,
            "signer_key": signer_key
        }
        if callback_url:
            data["callback_url"] = callback_url
        if callback_stages:
            data["callback_stages"] = callback_stages
        response = self.request_handler.post("/".join([factom_sdk.utils.consts.IDENTITY_URL, identity_chain_id,
                                                       factom_sdk.utils.consts.KEYS_STRING]), data)
        if new_public_key is None:
            response["key_pair"] = key_pair
        return response



import validators
import factom_sdk.utils.consts
from factom_sdk.utils.common_util import CommonUtil
from factom_sdk.utils.key_common import KeyCommon
from factom_sdk.request_handler.request_handler import RequestHandler
from factom_sdk.utils.identities.identities_key_util import IdentitiesKeyUtil


class IdentitiesClient:
    def __init__(self, base_url: str, app_id: str, app_key: str):
        self.request_handler = RequestHandler(base_url, app_id, app_key)
        self.keys = IdentitiesKeyUtil(base_url, app_id, app_key)

    def create(self, names: list, keys: list = None, callback_url: str = "", callback_stages: list = None):
        """Creates a new Identity chain.

        Args:
            names (list): The names array for your identity must be unique.
            keys (:obj:`list`, optional): An array of public key strings in base58 idpub format, ordered from the
            highest to the lowest priority.
            callback_url (:obj:`str`, optional): The URL where you would like to receive the callback from Connect.
            callback_stages (:obj:`list`, optional): The immutability stages you would like to be notified about.
            This list can include any or all of the three stages: `replicated`, `factom`, and `anchored`.

        Returns:
            Identity chain created object.
        """
        if callback_stages is None:
            callback_stages = []
        if not names:
            raise Exception("at least 1 name is required.")
        if not isinstance(names, list):
            raise Exception("names must be an array.")

        key_pairs = []
        if keys is not None:
            if not keys:
                raise Exception("at least 1 key is required.")
            if not isinstance(keys, list):
                raise Exception("keys must be an array.")
            signer_keys = keys
        else:
            key_pairs = [KeyCommon.create_key_pair() for _ in range(3)]
            signer_keys = [val["public_key"] for val in key_pairs]

        if not isinstance(callback_stages, list):
            raise Exception("callback_stages must be an array.")
        invalid_keys = KeyCommon.get_invalid_keys(signer_keys)
        if len(invalid_keys) > 0:
            raise Exception(invalid_keys)
        duplicate_keys = KeyCommon.get_duplicate_keys(signer_keys)
        if len(duplicate_keys) > 0:
            raise Exception(duplicate_keys)
        if callback_url and not validators.url(callback_url):
            raise Exception("callback_url is an invalid url format.")
        name_byte_count = 0
        for _name in names:
            name_byte_count += len(_name)

        # 2 bytes for the size of extid + 13 for actual IdentityChain ext-id text
        # 2 bytes per name for size * (number of names) + size of all names
        # 23 bytes for `{"keys":[],"version":1}`
        # 58 bytes per `"idpub2PHPArpDF6S86ys314D3JDiKD8HLqJ4HMFcjNJP6gxapoNsyFG",` array element
        # -1 byte because last key element has no comma
        # = 37 + name_byte_count + 2(number_of_names) + 58(number_of_keys)
        total_bytes = 37 + name_byte_count + (len(names) * 2) + (len(signer_keys) * 58)
        if total_bytes > 10240:
            message = "Entry size {} must be less than 10240. Use less/shorter names or less keys.".format(total_bytes)
            raise Exception(message)

        names_base64 = [CommonUtil.base64_encode(_name) for _name in names]
        data = {
            "names": names_base64,
            "keys": signer_keys
        }
        if callback_url:
            data["callback_url"] = callback_url
        if callback_stages:
            data["callback_stages"] = callback_stages
        response = self.request_handler.post(factom_sdk.utils.consts.IDENTITY_URL, data)
        if keys is None:
            response["key_pairs"] = key_pairs
        return response

    def get(self, identity_chain_id: str):
        """Gets a summary of the identity chain's current state.

        Args:
            identity_chain_id (str): The unique identifier for the identity chain being requested.

        Returns:
            Identity chain object.
        """
        if not identity_chain_id:
            raise Exception("identity_chain_id is required.")
        return self.request_handler.get("/".join([factom_sdk.utils.consts.IDENTITY_URL, identity_chain_id]))



import factom_sdk.utils.consts
from factom_sdk.request_handler.request_handler import RequestHandler


class AnchorsClient:
    def __init__(self, base_url: str, app_id: str, app_key: str):
        self.request_handler = RequestHandler(base_url, app_id, app_key)

    def get(self, object_hash: str = None, height: int = None):
        """Gets the anchors for a specific entry or directory block from Connect.

        Args:
            object_hash (str): The unique identifier for an entry or directory block to request an anchor for
            height (int): The height fort the directory block to request an anchor for

        Returns:
            Anchors object.
        """
        if object_hash is None:
            assert height is not None, "either object_hash or height must not be None"
            assert isinstance(height, int) and height > 0, "height must be a positive integer"
            return self.request_handler.get("/".join([factom_sdk.utils.consts.ANCHORS_URL, str(height)]))
        else:
            assert isinstance(object_hash, str) and len(object_hash) == 64, "object_hash must be a string of length 64"
            assert height is None, "object_hash provided, height must be None"
            return self.request_handler.get("/".join([factom_sdk.utils.consts.ANCHORS_URL, object_hash]))



import requests
import validators
from urllib.parse import urljoin
from factom_sdk.utils.common_util import CommonUtil


class RequestHandler:
    def __init__(self, base_url: str, app_id: str, app_key: str):
        if not validators.url(base_url):
            raise Exception("The base_url provided is not valid.")
        if not isinstance(app_id, str):
            raise Exception("The app_id provided is not valid.")
        if not isinstance(app_key, str):
            raise Exception("The app_key provided is not valid.")

        if not base_url.endswith("/"):
            base_url += "/"

        self.base_url = base_url
        self.app_id = app_id
        self.app_key = app_key

    def _generic_request(self, method, endpoint, data: dict = None, params: dict = None):
        headers = {
            "app_id": self.app_id,
            "app_key": self.app_key
        }
        response = requests.request(method, urljoin(self.base_url, endpoint), params=params, json=data, headers=headers)
        try:
            response.raise_for_status()
            return CommonUtil.decode_response(response.json())
        except requests.HTTPError as error:
            raise error

    def get(self, endpoint: str = "", params: dict = None):
        return self._generic_request("GET", endpoint, params=params)

    def post(self, endpoint: str = "", data: dict = None):
        return self._generic_request("POST", endpoint, data)



from factom_sdk.utils.key_common import KeyCommon


class Utils:
    @staticmethod
    def to_military_timezone_str(dt):
        return dt.isoformat().replace("+00:00", "Z")

    @staticmethod
    def generate_key_pair():
        """Creates a Public/Private Key Pair.

        Returns:
            A key pair object with Public and Private keys.
        """
        return KeyCommon.create_key_pair()



import factom_sdk.utils.consts
from factom_sdk.request_handler.request_handler import RequestHandler


class ReceiptsClient:
    def __init__(self, base_url: str, app_id: str, app_key: str):
        self.request_handler = RequestHandler(base_url, app_id, app_key)

    def get(self, entry_hash: str):
        """Gets a receipt for a specific entry from Connect.

        Args:
            entry_hash (str): The unique identifier created for the entry.

        Returns:
            Receipt object.
        """
        assert isinstance(entry_hash, str) and len(entry_hash) == 64, "entry_hash must be a string of length 64"

        return self.request_handler.get("/".join([factom_sdk.utils.consts.RECEIPTS_URL, entry_hash]))



CHAINS_URL = "chains"
IDENTITIES_URL = "identities"
KEYS_STRING = "keys"
SEARCH_URL = "search"
IDENTITY_URL = "identities"
ENTRIES_URL = "entries"
FIRST_URL = "first"
LAST_URL = "last"
RECEIPTS_URL = "receipts"
ANCHORS_URL = "anchors"
PRIVATE_PREFIX_BYTES = bytearray([0x03, 0x45, 0xf3, 0xd0, 0xd6])
PUBLIC_PREFIX_BYTES = bytearray([0x03, 0x45, 0xef, 0x9d, 0xe0])
BASE64_ENCODE = "base64"



from factom_sdk.client.chains_client import ChainsClient
from factom_sdk.client.identities_client import IdentitiesClient
from factom_sdk.client.receipts_client import ReceiptsClient
from factom_sdk.client.anchors_client import AnchorsClient
from factom_sdk.client.api_info_client import ApiInfoClient
from factom_sdk.utils.utils import Utils


class FactomClient:
    def __init__(self, base_url: str, app_id: str, app_key: str, automatic_signing: bool = True):
        """The Factom SDK is initialized.

        Args:
            base_url (str): The URL with the account domain
            app_id (str): Your app id
            app_id (str): Your app key
            automatic_signing (bool)

        Returns:
            Factom SDK client object.
        """
        if not isinstance(automatic_signing, bool):
            automatic_signing = True
        self.base_url = base_url
        self.app_id = app_id
        self.app_key = app_key
        self.automatic_signing = automatic_signing
        self._api_info = None
        self._chains = None
        self._identities = None
        self._receipts = None
        self._anchors = None
        self.utils = Utils

    @property
    def identities(self):
        if self._identities is None:
            self._identities = IdentitiesClient(self.base_url, self.app_id, self.app_key)
        return self._identities

    @property
    def api_info(self):
        if self._api_info is None:
            self._api_info = ApiInfoClient(self.base_url, self.app_id, self.app_key)
        return self._api_info

    @property
    def chains(self):
        if self._chains is None:
            self._chains = ChainsClient(self.base_url, self.app_id, self.app_key, self.automatic_signing)
        return self._chains

    @property
    def receipts(self):
        if self._receipts is None:
            self._receipts = ReceiptsClient(self.base_url, self.app_id, self.app_key)
        return self._receipts

    @property
    def anchors(self):
        if self._anchors is None:
            self._anchors = AnchorsClient(self.base_url, self.app_id, self.app_key)
        return self._anchors



import requests
import codecs
import factom_sdk.utils.consts
from factom_sdk.utils.key_common import KeyCommon


class ValidateSignatureUtil:
    @staticmethod
    def validate_signature(obj: dict, validate_for_chain: bool = True, request_handler=None):
        external_ids = obj["data"]["external_ids"]

        try:
            type_name = "SignedChain"
            invalid_format = "not_signed/invalid_chain_format"
            if not validate_for_chain:
                type_name = "SignedEntry"
                invalid_format = "not_signed/invalid_entry_format"
            # Hex signature always have length = 128.
            # And when convert it to base64, it always have length = 88
            # Replace "\n" to "" because codecs lib return "\n" in content
            if len(external_ids) < 6 \
                    or external_ids[0] != type_name \
                    or external_ids[1] != "0x01" \
                    or not KeyCommon.validate_checksum(external_ids[3]) \
                    or not (len(external_ids[4]) == 128
                            and len(
                        codecs.encode(codecs.decode(external_ids[4], "hex"), "base64").decode().replace("\n", "")
                    ) == 88):
                return invalid_format
        except ValueError:
            return invalid_format

        signer_chain_id = external_ids[2]
        signer_public_key = external_ids[3]
        signature = codecs.encode(codecs.decode(external_ids[4], "hex"), "base64").decode()
        time_stamp = external_ids[5]
        key_height = 0
        if "dblock" in obj["data"] \
                and obj["data"]["dblock"] is not None \
                and "height" in obj["data"]["dblock"]:
            key_height = obj["data"]["dblock"]["height"]
        try:
            key_response = request_handler.get("/".join([factom_sdk.utils.consts.IDENTITIES_URL, signer_chain_id,
                                                         factom_sdk.utils.consts.KEYS_STRING, signer_public_key]))
            if key_response["data"]["retired_height"] is not None and \
                    not ((key_response["data"]["activated_height"] <= key_height) and
                         (key_height <= key_response["data"]["retired_height"])):
                return "retired_key"
        except requests.HTTPError as error:
            if error.response.status_code == 404:
                return "key_not_found"
            raise error

        message = signer_chain_id + obj["data"]["content"] + time_stamp
        if not KeyCommon.validate_signature(signer_public_key, signature, message):
            return "invalid_signature"

        return "valid_signature"



import os
import hashlib
import base58
import ed25519
import base64
import factom_sdk.utils.consts


class KeyCommon:
    @staticmethod
    def create_key_pair():
        private_key_bytes = os.urandom(32)
        tmp = hashlib.sha256(hashlib.sha256(factom_sdk.utils.consts.PRIVATE_PREFIX_BYTES +
                                            private_key_bytes).digest()).digest()
        checksum = tmp[:4]
        private_key = base58.b58encode(bytes(factom_sdk.utils.consts.PRIVATE_PREFIX_BYTES +
                                             private_key_bytes + checksum))
        signing_key = ed25519.SigningKey(private_key_bytes)
        public_key_bytes = signing_key.get_verifying_key().to_bytes()
        tmp = hashlib.sha256(hashlib.sha256(factom_sdk.utils.consts.PUBLIC_PREFIX_BYTES +
                                            public_key_bytes).digest()).digest()
        checksum = tmp[:4]
        public_key = base58.b58encode(bytes(factom_sdk.utils.consts.PUBLIC_PREFIX_BYTES +
                                            public_key_bytes + checksum))
        return {
            "private_key": "".join(chr(x) for x in private_key),
            "public_key": "".join(chr(x) for x in public_key)
        }

    @staticmethod
    def validate_checksum(signer_key: str):
        if not signer_key:
            return False
        signer_key_bytes = base58.b58decode(signer_key)
        if len(signer_key_bytes) != 41:
            return False
        prefix_bytes = signer_key_bytes[:5]
        key_bytes = signer_key_bytes[5:37]
        checksum = signer_key_bytes[37:]
        tmp = hashlib.sha256(hashlib.sha256(prefix_bytes + key_bytes).digest()).digest()
        tmp_checksum = tmp[:4]
        if checksum != tmp_checksum:
            return False
        return True

    @staticmethod
    def get_invalid_keys(signer_keys: list):
        if not signer_keys:
            return []
        return [{"key": key, "error": "key is invalid"} for key in signer_keys if not KeyCommon.validate_checksum(key)]

    @staticmethod
    def get_duplicate_keys(signer_keys: list):
        if not signer_keys:
            return []

        seen = {}
        duplicates = []
        for key in signer_keys:
            if key not in seen:
                seen[key] = 1
            else:
                if seen[key] == 1:
                    duplicates.append({"key": key, "error": "key is duplicated, keys must be unique."})
                seen[key] += 1

        return duplicates

    @staticmethod
    def get_key_bytes_from_key(signer_key: str):
        if not KeyCommon.validate_checksum(signer_key):
            raise Exception("key is invalid.")
        signer_key_bytes = base58.b58decode(signer_key)
        return signer_key_bytes[5:37]

    @staticmethod
    def get_public_key_from_private_key(signer_private_key: str):
        if not KeyCommon.validate_checksum(signer_private_key):
            raise Exception("signer_private_key is invalid.")
        private_key_bytes = KeyCommon.get_key_bytes_from_key(signer_private_key)
        signing_key = ed25519.SigningKey(private_key_bytes)
        public_key_bytes = signing_key.get_verifying_key().to_bytes()
        tmp = hashlib.sha256(hashlib.sha256(factom_sdk.utils.consts.PUBLIC_PREFIX_BYTES +
                                            public_key_bytes).digest()).digest()
        checksum = tmp[:4]
        return "".join(chr(x) for x in base58.b58encode(bytes(factom_sdk.utils.consts.PUBLIC_PREFIX_BYTES +
                                                              public_key_bytes + checksum)))

    @staticmethod
    def sign_content(signer_private_key: str, message: str):
        if not signer_private_key:
            raise Exception("signer_private_key is required.")
        if not KeyCommon.validate_checksum(signer_private_key):
            raise Exception("signer_private_key is invalid.")
        if not message:
            raise Exception("message is required.")
        private_key_bytes = KeyCommon.get_key_bytes_from_key(signer_private_key)
        secret_key = ed25519.SigningKey(private_key_bytes)
        message_bytes = message.encode()
        return "".join(chr(x) for x in base64.b64encode(secret_key.sign(message_bytes)))

    @staticmethod
    def validate_signature(signer_public_key: str, signature: str, message: str):
        if not signer_public_key:
            raise Exception("signer_public_key is required.")
        if not KeyCommon.validate_checksum(signer_public_key):
            raise Exception("signer_public_key is invalid.")
        if not signature:
            raise Exception("signature is required.")
        if not message:
            raise Exception("message is required.")
        signature_bytes = base64.b64decode(signature)
        message_bytes = message.encode()
        key_bytes = KeyCommon.get_key_bytes_from_key(signer_public_key)
        verify_key = ed25519.VerifyingKey(key_bytes)
        try:
            verify_key.verify(signature_bytes, message_bytes)
            return True
        except ed25519.BadSignatureError:
            return False



from setuptools import setup, find_packages  # noqa: H301

NAME = "factom-harmony-connect"
VERSION = "1.0.3"
REQUIRES = [
    "base58 >= 1.0.3",
    "ed25519 >= 1.4",
    "validators >= 0.12",
    "requests >= 2.21",
]

setup(
    name=NAME,
    version=VERSION,
    author="Factom, Inc.",
    author_email="harmony-support@factom.com",
    description="Factom Harmony Connect SDK",
    long_description_content_type="text/markdown",
    url="https://github.com/FactomProject/factom-harmony-connect-python-sdk",
    keywords=["factom", "factom-blockchain", "blockchain", "blockchain-as-a-service", "SDK", "Harmony Connect"],
    install_requires=REQUIRES,
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    long_description="""
GETTING STARTED
===============

This section contains a summary of the steps required to get started
with Python Connect SDK installation.

System Requirements
-------------------

In order to use this Python SDK, you will need the following tool:

-  Python version >= 3.5


Installation
-------------
**Published package**

`pip install factom-harmony-connect`

**Open-source package**

- Clone the repo
- Setup a virtual environment (optional)
- Install dependencies
  - `pip install -r requirements.txt`

To use the SDK, you have to import: `from factom_sdk import FactomClient`

For robust documentation, see the github repository
[HERE](https://github.com/FactomProject/factom-harmony-connect-python-sdk)
""")



from factom_sdk.request_handler.request_handler import RequestHandler


class ApiInfoClient:
    def __init__(self, base_url: str, app_id: str, app_key: str):
        self.request_handler = RequestHandler(base_url, app_id, app_key)

    def get(self):
        """Gets general information about the Connect API.

        Returns:
            API Info object.
        """
        return self.request_handler.get()
